from python_speech_features import mfcc,delta,fbank
import numpy as np
from scipy.fftpack import fft,ifft
import ctypes as ct
import librosa

class FloatBits(ct.Structure):
    _fields_ = [
        ('M', ct.c_uint, 23),
        ('E', ct.c_uint, 8),
        ('S', ct.c_uint, 1)
    ]


class Float(ct.Union):
    _anonymous_ = ('bits',)
    _fields_ = [
        ('value', ct.c_float),
        ('bits', FloatBits)
    ]


def nextpow2(x):
    if x < 0:
        x = -x
    if x == 0:
        return 0
    d = Float()
    d.value = x
    if d.M == 0:
        return d.E - 127
    return d.E - 127 + 1

def MFCC(input,p=13,diff1=True,diff2=True,energy=True,frameDuration = 0.03, overlapRate = 0.5):
    """Calculate the MFCC features of the Audio. 

    input: An Audio object. 
    p: MFCC order. 
    frameDuration: A float object for the duration of each frame (seconds) or a int object for the length of each fram (sample points). 
    overlapRate: A float object in [0,1) for the overlapping rate of the frame. 
    diff1: Use first-order differential features. 
    diff2: Use second-order differential features. 
    energy: Use energy features. 

    return: A 2-D NumPy array of MFCC features. Each row will be MFCC features of one frame. 
    """
    if type(frameDuration) == float:
        windowLength=int(frameDuration*input.sr)
    else:
        windowLength=frameDuration
        frameDuration=windowLength/input.sr
    stepDuration=frameDuration*(1-overlapRate)
    wav_feature=mfcc(input.samples,input.sr,numcep=p,winlen = frameDuration, winstep=stepDuration, nfft=windowLength,appendEnergy=energy)
    feats = [wav_feature]
    if diff1:
        feats.append(delta(wav_feature, 1))
    if diff2:
        feats.append(delta(wav_feature, 2))
    return np.hstack(feats)

def fBank(input, filters = 26, frameDuration = 0.03, overlapRate = 0.5):
    """Calculate the Fbank features of the Audio. 

    input: An Audio object. 
    filters: Number of mel filters applied. 
    frameDuration: A float object for the duration of each frame (seconds) or a int object for the length of each fram (sample points). 
    overlapRate: A float object in [0,1) for the overlapping rate of the frame. 

    return: A 2-D NumPy array of FBank features. Each row will be fBank features of one frame. 
    """
    if type(frameDuration) == float:
        windowLength=int(frameDuration*input.sr)
    else:
        windowLength=frameDuration
        frameDuration=windowLength/input.sr
    stepDuration=frameDuration*(1-overlapRate)
    wav_feature,_=fbank(input.samples,input.sr,winlen = frameDuration, winstep=stepDuration, nfft=windowLength, nfilt=filters)
    return wav_feature

def melSpec(input, spec_h=128, spec_w=128):
    """Calculate the Mel spectrogram features of the Audio. 

    input: An Audio object. 
    spec_h: The height of the mel spectrogram features, which determines the frequency resolution.
    spec_w: The width of the mel spectrogram features, which determines the temporal resolution.

    return: A 2-D NumPy array of Mel spectrogram features. 
    """
    NUM_MELS = spec_h
    HOP_LENGTH = int(input.sr / (spec_w - 1)) 
    mel_spec = librosa.feature.melspectrogram(y=input.samples, sr=input.sr, hop_length=HOP_LENGTH, n_mels=NUM_MELS)
    mel_spec_db = librosa.power_to_db(mel_spec, ref=np.max)
    return mel_spec_db

def levinson(r, lpcOrder):
    a = np.zeros(lpcOrder + 1)
    e = np.zeros(lpcOrder + 1)
    a[0] = 1.0
    a[1] = - r[1] / r[0]
    e[1] = r[0] + r[1] * a[1]
    lam = - r[1] / r[0]
    for k in range(1, lpcOrder):
        lam = 0.0
        for j in range(k + 1):
            lam -= a[j] * r[k + 1 - j]
        lam /= e[k]
        U = [1]
        U.extend([a[i] for i in range(1, k + 1)])
        U.append(0)
        V = [0]
        V.extend([a[i] for i in range(k, 0, -1)])
        V.append(1)
        a = np.array(U) + lam * np.array(V)
        e[k + 1] = e[k] * (1.0 - lam * lam)
    return a, e[-1]

def myLPC(x,p):
    X = fft(x,2**nextpow2(len(x)))
    R = ifft(abs(X)**2)
    R = R/len(x)
    R = np.real(R)
    a, e  = levinson(R, p)
    a=np.real(a)
    return a,e

def LPC1time(input,p):
    x=input
    lpc,e=myLPC(np.asfortranarray(x), p)
    lpc=lpc[1:]
    return e,lpc

def LPC(input, p=10):
    """Matlab style LPC for each frame. 

    input: An AudioFrames object. 
    p: LPC order. 

    return: 
        List of LPC error of each frame. 
        List of LPC coefficient of each frame. 
    """
    y1=input.samples
    es=[]
    ws=[]
    for i in range(0,y1.shape[1]):
        x=np.squeeze(y1[:,i])
        e,lpc=LPC1time(np.asfortranarray(x), p)
        es.append(e)
        ws.append(lpc)
    return es,ws


from .freqDomainAnalyse import PSD, FFT, getMaxFrequency, specEntropy
from .timeDomainAnalyse import power, energy, zerocrossing, autocorr, snr
from .modelBasedAnalyse import MFCC, LPC, fBank, melSpec


import librosa
from .audio import Audio
import numpy as np

def read_Audio(direction):
    """Get audio file from specified path. Dual channel audio returns a tuple. 
    
    direction: The path.
    return:
        An Audio object when reading single channel audio.
        A tuple with two Audio objects when reading double channels audio. 
    """
    y,r=librosa.load(direction,sr=None)
    if len(y.shape)==2:
        return (Audio(y[0],r),Audio(y[1],r))
    elif len(y.shape)==1:
        return Audio(y,r)
    else:
        raise Exception("Wrong number of tracks. ")


def save_Audio_DoubleTracks(audio1,audio2,direction):
    """Combine two audio with the same length and sampling rate into a dual channel audio and save it to the specified path.
    
    audio1: An Audio object for the first channel. 
    audio2: An Audio object for the second channel. 
    direction: The saving path. 
    """
    if audio1.sr!=audio2.sr:
        raise Exception("Two audios should have the same sample rate. ")
    librosa.output.write_wav(direction,np.asfortranarray(np.array([np.real(audio1.samples),np.real(audio2.samples)])),audio1.sr)


import numpy as np
from math import *

class axis:
    def __init__(self, length, sr, int_index_map, float_index_map, index_map, scales_map):
        self.scale = np.arange(0, length, 1)
        self.sr = sr
        self.default_start = 0
        self.default_end = self.scale.shape[0]
        self.default_step = 1
        self.int_index_map = int_index_map
        self.float_index_map = float_index_map
        self.index_map = index_map
        self.scales_map = scales_map
    
    def __getitem__(self, index):
        if type(index) == slice:
            start = index.start
            end = index.stop
            step = index.step
            if start is None:
                start = self.default_start
            elif type(start) == float:
                start = self.float_index_map(start)
            else:
                start = self.int_index_map(start)
            if end is None:
                end = self.default_end
            elif type(end) == float:
                end = self.float_index_map(end)
            else:
                end = self.int_index_map(end)
            if step is None:
                step = self.default_step
            elif type(step) == float:
                step = self.float_index_map(step)
            else:
                step = self.int_index_map(step)
            return self.scale[start:end:step]
        else:
            if type(index) == float:
                index = self.float_index_map(index)
            else:
                index = self.int_index_map(index)
            return self.scale[index]
    
    def slicing(self, start, end, step, index_type):
        start = self.index_map[index_type](start)
        if end is None:
            end = self.default_end
        else:
            end = self.index_map[index_type](end)
        if step is None:
            step = self.default_step
        else:
            step = self.index_map[index_type](step)
        return self.scale[start:end:step]
            
    def getScale(self, index_type):
        return self.scales_map[index_type](self.scale)

class time_axis(axis):
    def __init__(self, length, sr):
        super().__init__(length, sr, 
        lambda x:x, 
        lambda x: int(self.sr * x), 
        {"t/s": lambda x: int(self.sr * x), "t/ms": lambda x: int(self.sr * x / 1000), "n": lambda x:x}, 
        {"n": lambda x: x, "t/s": lambda x: x / sr, "t/ms": lambda x: x * 1000 / sr})

class time_delta_axis(axis):
    def __init__(self, length, sr):
        max_diff = int((length + 1) / 2) - 1
        super().__init__(length, sr, 
        lambda x:x + max_diff, 
        lambda x: int(self.sr * x + max_diff), 
        {"t/s": lambda x: int(self.sr * x + max_diff), "t/ms": lambda x: int(self.sr * x / 1000 + max_diff), "k": lambda x:x + max_diff}, 
        {"t/s": lambda x: (x - max_diff)/sr, "t/ms": lambda x: 1000*(x - max_diff)/sr, "k": lambda x: x - max_diff})

class frame_axis(axis):
    def __init__(self, length, sr, step):
        time_step = step / sr
        super().__init__(length, sr, 
        lambda x: int(x / step), 
        lambda x: int(x / time_step), 
        {"t/s": lambda x: int(x / time_step), "t/ms": lambda x: int(x / time_step / 1000), "n": lambda x: int(x / step), "frame": lambda x:x}, 
        {"t/s": lambda x: x * time_step, "t/ms": lambda x: 1000 * x * time_step, "n": lambda x: x * step, "frame": lambda x:x})

class freq_axis(axis):
    def __init__(self, length, sr):
        super().__init__(length, sr, 
        lambda x: x, 
        lambda x: int(x/sr*length), 
        {"frequency/Hz": lambda x: int(x/sr*length), "frequency/(rad/s)": lambda x: int(x/sr*length/2/np.pi), "normalized frequency/Hz": lambda x:int(x*length), "normalized frequency/(rad/s)": lambda x:int(x*length/np.pi/2), "freq point": lambda x:x}, 
        {"frequency/Hz": lambda x: x*sr/length, "frequency/(rad/s)": lambda x: x*sr/length*2*np.pi, "normalized frequency/Hz": lambda x:x/length, "normalized frequency/(rad/s)": lambda x:2*x/length*np.pi, "freq point": lambda x:x})
        self.default_end = int(length / 2)
    
    def melSlicing(self, start, end, step, index_type):
        start = self.index_map[index_type](start)
        if end is None:
            end = self.default_end
        else:
            end = self.index_map[index_type](end)
        if step is None:
            step = self.default_step
        else:
            step = self.index_map[index_type](step)
        
        steps = int((end - start) / step)
        start = self.scales_map["frequency/Hz"](start)
        end = self.scales_map["frequency/Hz"](end)

        slicing = 700 * (np.power(10, np.linspace(2595 * np.log10(1 + start/700), 2595 * np.log10(1 + end/700), steps)/2595) - 1)
        slicing = np.array([self.index_map["frequency/Hz"](s) for s in slicing.tolist()])
        return self.scale[slicing.astype(int)]
    
    def logSlicing(self, start, end, step, index_type):
        start = self.index_map[index_type](start)
        if end is None:
            end = self.default_end
        else:
            end = self.index_map[index_type](end)
        if step is None:
            step = self.default_step
        else:
            step = self.index_map[index_type](step)
        
        steps = int((end - start) / step)

        slicing = np.power(10, np.linspace(np.log10(1 + start), np.log10(1 + end), steps)) - 1
        return self.scale[slicing.astype(int)]


from pyAudioKits.audio.audio import typeJudge
from scipy.signal import lfilter, filtfilt, lfilter_zi

def ltiFilter(input,numerators,denominators,zero_phase=False):
    """LTI filter design by specifying the denominator and numerator coefficients of the system function. 

    input: An Audio object or an Audio frame object. 
    numerators: An Numpy array of the numerator coefficients of the system function. 
    denominators: An Numpy array of the denominator coefficients of the system function. 
    zero_phase: Use bi-directional filtering to maintain a phase response of 0. 
    return: 
        An Audio object if the input is an Audio object. 
        An AudioFrames object if the input is an AudioFrames object. 
    """
    if zero_phase:
        y=filtfilt(numerators,denominators,input.samples, axis=-1)
    else:
        y,_=lfilter(numerators,denominators,input.samples, axis=-1,zi=input.samples[-1]*lfilter_zi(numerators,denominators))
    return typeJudge(input,y)


from .audio import Audio, concatenate, synthesis
from .io import read_Audio, save_Audio_DoubleTracks
from .simulate import create_Single_Freq_Audio, mixWithSNR
from .record import start_record, end_record, record


from scipy.fftpack import fft,ifft
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
import pandas as pd
from pyAudioKits.datastructures.ds import tResult, fResult
import librosa

def calculate_H(frame_w,N):
    freq = FFT(frame_w,N).samples
    Y = freq * np.conj(freq)
    sumY=np.sum(Y,axis=-1)
    if type(sumY) == np.ndarray:
        P = Y/sumY[:,None]
    else:
        P = Y/sumY
    H=-np.sum(P*(np.log2(P+1e-6)),axis=-1)
    return H

def specEntropy(input,N=None):
    """Calculate the spectral entropy of the audio. 

    input: An Audio object or an AudioFrames object. 
    N: Num of Fourier Transform points. None if use all samples. 
    return: 
        If input is an Audio object: A float object for the spectral entropy of the whole audio. 
        If input is an AudioFrames object: A frame_axis_ds object containing spectral entropy of each frame. 
    """
    if N is None:
        N=input.samples.shape[-1]
    return tResult(calculate_H(input,N), input, "bit")

def PSD(input,N=None,dB=False):
    """Calculate the power spectral density of the audio. 

    input: An Audio object or an AudioFrames object. 
    N: Num of Fourier Transform points. None if use all samples. 
    dB: Whether to express the output in gain (dB). 
    return: 
        If input is an Audio object: A freq_axis_ds_real object containing power spectral density of the whole audio. 
        If input is an AudioFrames object: A freq_frame_axis_ds_real object containing power spectral density of each frame. 
    """
    if N is None:
        N=input.samples.shape[-1]
    freq=fft(input.samples,N,axis=-1)
    result = np.power(np.abs(freq),2)/N
    if dB:
        result = 10 * np.log10(result)
        unit = "dB"
    else:
        unit = "amplitude"
    return fResult(result, input, unit)

def FFT(input,N=None):
    """Calculate the FFT of the audio. 

    input: An Audio object or an AudioFrames object (STFT). 
    N: Num of Fourier Transform points. None if use all samples. 
    return: 
        If input is an Audio object: A freq_axis_ds object containing spectrum. 
        If input is an AudioFrames object: A freq_frame_axis_ds object containing short-time spectrum. 
    """
    if N is None:
        N=input.samples.shape[-1]
    result = fft(input.samples,n=N,axis=-1)
    return fResult(result,input,"complex")

def getMaxFrequency(input,N=None,dB=False):
    """Get the frequency and the amplitude of spectrum peak. 

    input: An Audio object or an AudioFrames object. 
    N: Num of Fourier Transform points. None if use all samples. 
    dB: Whether to express the output amplitude in gain(dB). 
    return: 
        If input is an Audio object: The frequency and the amplitude of the spectrum peak of the whole audio. 
        If input is an AudioFrames object: A frame_axis_ds object containing the frequency of the spectrum peak and a frame_axis_ds object containing the amplitude of the spectrum peak. 
    """
    if N is None:
        N=input.samples.shape[-1]
    freq=FFT(input, N).samples
    indices = np.argmax(2*np.abs(freq), axis = -1)
    max_freq = freq[indices] if len(freq.shape) == 1 else freq[(np.matrix([np.arange(len(indices))]), np.matrix([indices]))][0]
    result = 2*np.abs(max_freq)/freq.shape[-1]
    if dB:
        result = 10 * np.log10(result)
        unit = "dB"
    else:
        unit = "amplitude"
    return tResult(indices*(input.sr/freq.shape[-1]), input,"Hz"), tResult(result, input,unit)


import numpy as np
import scipy
import scipy.signal
from pyAudioKits.datastructures.ds import time_delta_frame_axis_ds, time_delta_axis_ds, tResult
import librosa
        
def energyCal(frame_w):
    return np.sum(np.power(frame_w,2),axis=-1)

def overzeroCal(frame_w,fs):
    sgs=np.sign(frame_w.T)
    overzeros=np.sum(np.abs(sgs[1:]-sgs[:-1]),axis=0)
    return overzeros

def power(input, dB = False):
    """Calculate the power of the audio. 

    input: An Audio object or an AudioFrames object. 
    dB: Whether to express the result in the form of gain (dB). 
    return: 
        If input is an Audio object: Power of the whole audio. 
        If input is an AudioFrames object: A frame_axis_ds object containing power of each frame. 
    """
    result = np.sum(input.samples**2,axis=-1)/input.samples.shape[-1]
    if dB:
        result = 10 * np.log10(result)
        unit = "dB"
    else:
        unit = "amplitude"
    return tResult(result, input, unit)

def energy(input, dB = False):
    """Calculate the energy of the audio. 

    input: An Audio object or an AudioFrames object. 
    dB: Whether to express the result in the form of gain (dB). 
    return: 
        If input is an Audio object: Energy of the whole audio. 
        If input is an AudioFrames object: A frame_axis_ds object containing energy of each frame. 
    """
    result = energyCal(input.samples)
    if dB:
        result = 10 * np.log10(result)
        unit = "dB"
    else:
        unit = "amplitude"
    return tResult(result, input, unit)

def zerocrossing(input):
    """Calculate the zero crossing rate of the audio. 

    input: An AudioFrames object. 
    return: A frame_axis_ds object containing zero crossing rate of each frame. 
    """
    return tResult(overzeroCal(input.samples,input.sr), input, "Hz")
    
def autocorr(input):
    """Calculate the auto correlation function of the audio. 

    input: An Audio object or an AudioFrames object. 
    return: 
        If input is an Audio object: A time_delta_axis_ds object containing auto correlation function.
        If input is an AudioFrames object: A time_delta_frame_axis_ds object containing short-time auto correlation result for each frames.
    """
    if len(input.shape) == 2:
        corr = np.array([scipy.signal.correlate(input.samples[i],input.samples[i],"full") for i in range(input.shape[0])])
    else:
        corr = scipy.signal.correlate(input.samples,input.samples,"full")
    #corr = np.correlate(input.samples,input.samples,"full")
    if len(corr.shape) == 2:
        return time_delta_frame_axis_ds(corr, input.sr, input.step, "value")
    else:
        return time_delta_axis_ds(corr, input.sr, "value")

def snr(signal,signalAndNoise):
    """Calculate the SNR(dB). 
    
    signal: An Audio object. The ground truth of the signal. 
    signalAndNoise: An Audio object. Signal mixed with noise. 
    return: A float object of the SNR(dB). 
    """
    noise=signalAndNoise-signal
    signalPower=power(signal)
    noisePower=power(noise)
    return 10*np.log10(signalPower/noisePower)